# -*- coding: utf-8 -*-

"""Client library for interacting with the SIS RESTful API"""

__version__ = (1, 1, 0)
__author__ = 'Anton Gavrik'

import json

import logging
try:  # Python 2.7+
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass

LOG = logging.getLogger(__name__)
LOG.addHandler(NullHandler())


class Response(object):
    """Represent the client's response.

    Depending on the calling method behaves either like a dictionary 
    for methods returning a single object e.g. .get() or like a list
    for methods returning multiple objects e.g. .fetch_all()

    """
    def __init__(self, result, meta=None):
        self._result = result    
        self._meta = meta 

    def __len__(self):
        return len(self._result)

    def __getitem__(self, key):
        return self._result[key]

    def __setitem__(self, key, value):
        self._result[key] = value

    def __delitem__(self, key):
        del self._result[key]

    def __iter__(self):
        return iter(self._result)

    def __contains__(self, item):
        """x in Response 

        """     
        if item in self._result:
            return True
        else:
            return False
        
    def __str__(self):
        return str(self._result)

    def to_dict(self):
        """Returns a dict representaton of the object
        
        """
        if not isinstance(self._result, dict):
            raise Error('{0} is not a dict-like object'.format(
                self.__class__.__name__))
        return dict(self._result)

    def to_list(self):
        """Returns a list representation of the object
        
        """
        if not isinstance(self._result, list):
            raise Error('{0} is not a list-like object'.format(
                self.__class__.__name__))
        return list(self._result)


class Meta(object):
    """Represents meta data in the clients response."""

    def __init__(self, headers):
        """    
        args:
            headers: dict containing http headers

        Derived attributes:
            .total_count
                is created and is set to the value of 'x-total-count' header 
                if it's present. 
        """           
        self.headers = headers

        if 'x-total-count' in self.headers:
            self.total_count = int(self.headers['x-total-count'])


class Error(Exception):
    """SIS Error"""

    def __init__(self, error, http_status_code=None, 
                 code=None, response_dict={}):
        """
        args:
            error: string representing error description
            code: string representing error code
            http_status_code: HTTP status code
            response_dict: dictionary representing encoded 
                http response body 
        """
        self.error = error
        self.http_status_code = http_status_code
        self.code = code
        self.response_dict = response_dict

        super(Error, self).__init__(self.__str__())

    def __repr__(self):
        return ('Error(http_status_code="{http_status_code}", '
                'error={error}, code="{code}", '
                'response_dict={response_dict})'
                .format(error=json.dumps(self.error),
                        code=self.code,
                        http_status_code=self.http_status_code,
                        response_dict=json.dumps(str(self.response_dict))))

    def __str__(self):
        return self.error


from .client import Client




# -*- coding: utf-8 -*-

import logging
import json

from . import Error, Response, http, NullHandler

LOG = logging.getLogger(__name__)
LOG.addHandler(NullHandler())


class Endpoint(object):

    """SIS endpoint"""

    def __init__(self, endpoint, client):
        self.endpoint = endpoint
        self.client = client

    def fetch_page(self, query=None):
        """Returns a Response() list-like object

        Response()._meta.total_count is the total number of items

        """
        return self._get(query=query)

    def fetch_all(self, query=None):
        """Calls fetch_page() multiple times to retrieve all items,
        returns a Response() list-like object of items fetched.

        Response._meta.headers is set to headers of the last HTTP request

        """    
        if not query:
            query = {}

        results = []
        while True:
            response = self.fetch_page(query)
            results.extend(list(response))
            if len(results) >= response._meta.total_count:
                break
            query['offset'] = len(results)

        response._result = results
        return response

    def get(self, id):
        """API GET """
        return self._get(id)

    def create(self, content):
        """API POST """

        headers = self._get_headers(add_content=True)
        request = http.Request(uri=self._get_uri(),
                               method='POST',
                               body=json.dumps(content),
                               headers=headers)

        return self.client.request(request)

    def update(self, id, content, query=None):
        """API PUT """

        headers = self._get_headers(add_content=True)
        request = http.Request(uri=self._get_uri(id, query),
                               method='PUT',
                               body=json.dumps(content),
                               headers=headers)

        return self.client.request(request)

    def update_bulk(self, content, query=None):
        """API Bulk update.

        Returns: a Response dict-like object in the form of
        {
            'errors': [<items>],
            'success': [<items>]
        }
        """

        if isinstance(content, list):
            # Handle update if we're provided a list of changes
            headers = self._get_headers(add_content=True)
            request = http.Request(uri=self._get_uri(query=query),
                                   method='PUT',
                                   body=json.dumps(content),
                                   headers=headers)

            return self.client.request(request)

        elif isinstance(content, dict):
            # Handle update if we're provided a change and a query
            if query == None:
                err_msg = 'dictionary updates must be accompanied with a query'
                raise Error(http_status_code=400,
                            error=err_msg,
                            code=0,
                            response_dict={})

            headers = self._get_headers(add_content=True)
            request = http.Request(uri=self._get_uri(query=query),
                                   method='put',
                                   body=json.dumps(content),
                                   headers=headers)

            return self.client.request(request)

        else:
            err_msg = 'content must be a list of entities or an update dict' + \
                      ' with a query'
            raise Error(http_status_code=400,
                        error=err_msg,
                        code=0,
                        response_dict={ })


    def delete_bulk(self, query):
        """Bulk delete.

        Returns: a Response dict-like object in the form of
        {
            'errors': [<items>],
            'success': [<items>]
        }

        """
        if not isinstance(query, dict) \
            or not isinstance(query.get('q', None), dict):
            err_msg = 'query must be a dictionary and contain a "q" dict'
            raise Error(http_status_code=400,
                        error=err_msg,
                        code=0,
                        response_dict={ })

        q = query.get('q')
        if not len(list(q.keys())):
            err_msg = 'query dictionary must not be empty' 
            raise Error(http_status_code=400,
                        error=err_msg,
                        code=0,
                        response_dict={ })

        headers = self._get_headers(add_content=True)
        request = http.Request(uri=self._get_uri(query=query),
                               method='DELETE',
                               headers=headers)

        return self.client.request(request)

    def delete(self, id):
        """API DELETE"""

        headers = self._get_headers(add_content=True)
        request = http.Request(uri=self._get_uri(id),
                               method='DELETE',
                               headers=headers)

        return self.client.request(request)

    def _get(self, obj=None, query=None):
        headers = self._get_headers(add_content=True)
        request = http.Request(uri = self._get_uri(obj, query),
                               headers=headers)
        
        return self.client.request(request)

    def _get_headers(self, add_content):
        headers = {
            'Accept': 'application/json'
        }

        if add_content:
            headers['Content-Type'] = 'application/json'

        if self.client.auth_token:
            headers['x-auth-token'] = self.client.auth_token

        return headers

    def _get_uri(self, obj=None, query=None):
        query_str = ''
        path_str = ''

        # ignore empty dict
        if isinstance(query, dict) and query:
            query = query.copy()
            for k in query:
                # support cas / q params as dict
                if isinstance(query[k], dict):
                    query[k] = json.dumps(query[k])

            query_str = '?{0}'.format(http.urlencode(query))

        if obj:
            path_str = '/{0}'.format(str(obj))

        resp = '{0}/{1}{2}{3}'.format(self.client.base_uri, self.endpoint,
                                  path_str, query_str)
        return resp




# -*- coding: utf-8 -*-

import logging
import sys
import json

from . import Response, Error, Meta, NullHandler

LOG = logging.getLogger(__name__)
LOG.addHandler(NullHandler())

# attempt to use requests v2+(must have Session.prepare_request() method) 
# if available, else use the stdlib
HTTP_LIB = 'stdlib'
try:
    import requests
except ImportError:
    pass
else:
    try:
        if hasattr(requests.Session, 'prepare_request'):
            HTTP_LIB = 'requests'
    except AttributeError:
        pass

LOG.debug('using {0} to handle http'.format(HTTP_LIB))

if HTTP_LIB == 'stdlib':
    # for py versions 2.7.9+/3.4.3+ we require "unverified SSL context"
    # in order to disable SSL cert validation
    SSL_CONTEXT = None
    import ssl
    try:
        SSL_CONTEXT = ssl._create_unverified_context()
    except AttributeError:
        pass

    # py3
    if sys.version_info[0] >= 3:
        import urllib.request, urllib.error
        stdlib_request = urllib.request.Request
        stdlib_HTTPError = urllib.error.HTTPError   
        stdlib_urlopen = urllib.request.urlopen

    # py2
    else:
        import urllib
        import urllib2
        stdlib_request = urllib2.Request
        stdlib_HTTPError = urllib2.HTTPError
        stdlib_urlopen = urllib2.urlopen

elif HTTP_LIB == 'requests':
    # this will disable InsecureRequestWarning
    try:
        requests.packages.urllib3.disable_warnings()
    except AttributeError:
        pass

# urlencode method
if sys.version_info[0] >= 3:
    import urllib.parse
    urlencode = urllib.parse.urlencode
else:
    import urllib
    urlencode = urllib.urlencode


def get_handler(http_keep_alive=True):
    """Returns an appropriate http handler object based on the available 
    http library.

    """
    if HTTP_LIB == 'stdlib':    
        return StdLibHandler()

    elif HTTP_LIB == 'requests':
        return RequestsHandler(http_keep_alive=http_keep_alive)


class Request(object):

    """HTTP request proxy"""

    def __init__(self, uri, method='GET', body=None, headers=None):
        self.uri = uri
        self.method = method
        self.body = body
        self.headers = headers

    def __str__(self):
        s = ''
        s += '{0} '.format(self.method)
        s += '{0} '.format(self.uri)
        return s


class BaseHTTPHandler(object):

    """HTTP Handler proxy base class"""

    def __init__(self):
        pass

    def request(self, request):
        raise NotImplementedError


class StdLibHandler(BaseHTTPHandler):

    """Handles http using the standard library"""

    def __init__(self, *args, **kwargs): 
        super(StdLibHandler, self).__init__(*args, **kwargs)

    def request(self, request):
        # encode request.body (py3)
        # POST data should be bytes or an iterable of bytes.
        if request.body:
            request.body = request.body.encode('utf-8')

        # create Request() object, set uri and body contents if any
        new_req = stdlib_request(request.uri, data=request.body)

        # add headers if present
        if request.headers:
            for header_name in request.headers:
                new_req.add_header(header_name, request.headers[header_name])

        # set method, if different from GET
        if request.method != 'GET':
            # urllib method needs to be a callable method
            new_req.get_method = lambda: request.method

        # send request
        LOG.debug(request)
        try:
            # if SSL_CONTEXT is set urlopen is assumed to support context arg
            if SSL_CONTEXT:
                response = stdlib_urlopen(new_req, context=SSL_CONTEXT)
            else:
                response = stdlib_urlopen(new_req)

        except stdlib_HTTPError as e:
            # read response
            response_str = e.read().decode('utf-8')

            # decode response, trap non-json responses
            try:
                response_dict = json.loads(response_str)
            except ValueError:
                raise Error(http_status_code=e.code,
                            error=('Failed to decode JSON from the response: {0}'
                                   .format(response_str[:256].encode('utf-8'))))
                                                            
            code = response_dict.get('code')

            # lookup error in the response body,
            # if not available use http error info
            error = response_dict.get('error')
            if not error:
                error = e.reason

            raise Error(http_status_code=e.code,
                        error=error,
                        code=code,
                        response_dict=response_dict)

        # read response
        response_str = response.read().decode('utf-8')

        # decode response, trap non-json responses
        try:
            result = json.loads(response_str)
        except ValueError:
            raise Error(http_status_code=response.getcode(),
                        error=('Failed to decode JSON from the response: {0}'
                               .format(response_str[:256].encode('utf-8'))))

        # build meta with headers as a dict
        # py3
        if sys.version_info[0] >= 3:
            # python2.6 does not support dict comprehensions
            d = {}
            for k, v in response.info().items():
                d[k] = v
            meta = Meta(d)

        # py2
        else:
            meta = Meta(response.info().dict)

        # return Response object    
        return Response(result, meta)


class RequestsHandler(BaseHTTPHandler):

    """Handles HTTP using requests library"""          

    def __init__(self, http_keep_alive=True, *args, **kwargs):
        super(RequestsHandler, self).__init__(*args, **kwargs)

        self.http_keep_alive = http_keep_alive

        self._session = requests.Session()
        
    def request(self, request):
        LOG.debug(request)

        # add "Connection: close" header if not http_keep_alive 
        if not self.http_keep_alive:
            request.headers['Connection'] = 'close'

        # prepare request
        req = requests.Request(request.method, request.uri, 
                               data=request.body, headers=request.headers)
        prepped = self._session.prepare_request(req)

        # send request
        # stream=True immediately download the response 
        # content(default is False)
        # verify=False do not verify SSL cert
        response = self._session.send(prepped, stream=True, verify=False)

        # Explicitly set response encoding to 'utf-8'. This fixes long responses issue.
        # This is consistent with StdLib encoding handling.
        # Long responses issue:
        # https://github.com/requests/requests/issues/2359
        response.encoding = 'utf-8'

        # decode response, trap non-json responses
        try:
            response_dict = json.loads(response.text)
        except ValueError:
            raise Error(http_status_code=response.status_code,
                        error=('Failed to decode JSON from the response: {0}'
                               .format(response.text[:256].encode('utf-8'))))

        # raise Error if we got http status code >= 400
        if response.status_code >= 400:
            # lookup error and code in the response dict
            error = response_dict.get('error')
            code = response_dict.get('code')

            raise Error(http_status_code=response.status_code,
                        error=error,
                        code=code,
                        response_dict=response_dict)

        # create Meta
        meta = Meta(response.headers)
        
        # return Response()
        return Response(response_dict, meta)




# -*- coding: utf-8 -*-

import base64
import logging

from . import http, endpoint, NullHandler

LOG = logging.getLogger(__name__)
LOG.addHandler(NullHandler())


class Client(object):

    """SIS client"""

    def __init__(self, url, version=1.1, auth_token=None,
                 http_keep_alive=True):

        self.version = version
        self.base_uri = '{0}/api/v{1}'.format(url.rstrip('/'), self.version)
        self.auth_token = auth_token

        # get http handler
        self._http_handler = http.get_handler(http_keep_alive=http_keep_alive)

        # api endpoints
        self.schemas = endpoint.Endpoint('schemas', self)
        self.hooks = endpoint.Endpoint('hooks', self)
        self.hiera = endpoint.Endpoint('hiera', self)
        self.users = endpoint.Endpoint('users', self)

    def entities(self, schema_name):
        return endpoint.Endpoint('entities/{0}'.format(schema_name), self)

    def tokens(self, username):
        return endpoint.Endpoint('users/{0}/tokens'.format(username), self)

    def request(self, request):
        return self._http_handler.request(request)

    def authenticate(self, username, password):
        uri = '{0}/users/auth_token'.format(self.base_uri)

        # py3 base64.b64encode expects and returns a byte str
        enc_creds = base64.b64encode(
            '{0}:{1}'.format(username, password).encode('utf-8')).decode('utf-8')

        headers = { 'Authorization': 'Basic {0}'.format(enc_creds) }

        request = http.Request(uri=uri,
                               method='POST',
                               headers=headers)

        self.auth_token =  self.request(request)['name']

        return True




# -*- coding: utf-8 -*-
from distutils.core import setup

import sispy

version = '.'.join([ str(sispy.__version__[i]) for i in range(3) ])

setup (
    name='sispy',
    version=version,
    author=sispy.__author__,
    description='Python client library for interacting with the SIS RESTful API',
    packages=['sispy'],
    url='https://github.com/sis-cmdb/sis-python',
    download_url='https://github.com/sis-cmdb/sis-python/tarball/v%s' % version,
    license='BSD 3-Clause'
)

